@@ -25,6 +25,18 @@
 
 #include "ext2fs.h"
 
+static errcode_t read_ea_inode_hash(ext2_filsys fs, ext2_ino_t ino, __u32 *hash)
+{
+	struct ext2_inode inode;
+	errcode_t retval;
+
+	retval = ext2fs_read_inode(fs, ino, &inode);
+	if (retval)
+		return retval;
+	*hash = ext2fs_get_ea_inode_hash(&inode);
+	return 0;
+}
+
 #define NAME_HASH_SHIFT 5
 #define VALUE_HASH_SHIFT 16
 
@@ -46,7 +58,7 @@ __u32 ext2fs_ext_attr_hash_entry(struct ext2_ext_attr_entry *entry, void *data)
 	}
 
 	/* The hash needs to be calculated on the data in little-endian. */
-	if (entry->e_value_block == 0 && entry->e_value_size != 0) {
+	if (entry->e_value_inum == 0 && entry->e_value_size != 0) {
 		__u32 *value = (__u32 *)data;
 		for (n = (entry->e_value_size + EXT2_EXT_ATTR_ROUND) >>
 			 EXT2_EXT_ATTR_PAD_BITS; n; n--) {
@@ -59,6 +71,85 @@ __u32 ext2fs_ext_attr_hash_entry(struct ext2_ext_attr_entry *entry, void *data)
 	return hash;
 }
 
+/*
+ * ext2fs_ext_attr_hash_entry2()
+ *
+ * Compute the hash of an extended attribute.
+ * This version of the function supports hashing entries that reference
+ * external inodes (ea_inode feature).
+ */
+errcode_t ext2fs_ext_attr_hash_entry2(ext2_filsys fs,
+				      struct ext2_ext_attr_entry *entry,
+				      void *data, __u32 *hash)
+{
+	*hash = ext2fs_ext_attr_hash_entry(entry, data);
+
+	if (entry->e_value_inum) {
+		__u32 ea_inode_hash;
+		errcode_t retval;
+
+		retval = read_ea_inode_hash(fs, entry->e_value_inum,
+					    &ea_inode_hash);
+		if (retval)
+			return retval;
+
+		*hash = (*hash << VALUE_HASH_SHIFT) ^
+			(*hash >> (8*sizeof(*hash) - VALUE_HASH_SHIFT)) ^
+			ea_inode_hash;
+	}
+	return 0;
+}
+
+#undef NAME_HASH_SHIFT
+#undef VALUE_HASH_SHIFT
+
+#define BLOCK_HASH_SHIFT 16
+
+/* Mirrors ext4_xattr_rehash() implementation in kernel. */
+void ext2fs_ext_attr_block_rehash(struct ext2_ext_attr_header *header,
+				  struct ext2_ext_attr_entry *end)
+{
+	struct ext2_ext_attr_entry *here;
+	__u32 hash = 0;
+
+	here = (struct ext2_ext_attr_entry *)(header+1);
+	while (here < end && !EXT2_EXT_IS_LAST_ENTRY(here)) {
+		if (!here->e_hash) {
+			/* Block is not shared if an entry's hash value == 0 */
+			hash = 0;
+			break;
+		}
+		hash = (hash << BLOCK_HASH_SHIFT) ^
+		       (hash >> (8*sizeof(hash) - BLOCK_HASH_SHIFT)) ^
+		       here->e_hash;
+		here = EXT2_EXT_ATTR_NEXT(here);
+	}
+	header->h_hash = hash;
+}
+
+#undef BLOCK_HASH_SHIFT
+
+__u32 ext2fs_get_ea_inode_hash(struct ext2_inode *inode)
+{
+	return inode->i_atime;
+}
+
+void ext2fs_set_ea_inode_hash(struct ext2_inode *inode, __u32 hash)
+{
+	inode->i_atime = hash;
+}
+
+__u64 ext2fs_get_ea_inode_ref(struct ext2_inode *inode)
+{
+	return ((__u64)inode->i_ctime << 32) | inode->osd1.linux1.l_i_version;
+}
+
+void ext2fs_set_ea_inode_ref(struct ext2_inode *inode, __u64 ref_count)
+{
+	inode->i_ctime = (__u32)(ref_count >> 32);
+	inode->osd1.linux1.l_i_version = (__u32)ref_count;
+}
+
 static errcode_t check_ext_attr_header(struct ext2_ext_attr_header *header)
 {
 	if ((header->h_magic != EXT2_EXT_ATTR_MAGIC_v1 &&
@@ -69,9 +160,6 @@ static errcode_t check_ext_attr_header(struct ext2_ext_attr_header *header)
 	return 0;
 }
 
-#undef NAME_HASH_SHIFT
-#undef VALUE_HASH_SHIFT
-
 errcode_t ext2fs_read_ext_attr3(ext2_filsys fs, blk64_t block, void *buf,
 				ext2_ino_t inum)
 {
@@ -207,17 +295,19 @@ errcode_t ext2fs_adjust_ea_refcount(ext2_filsys fs, blk_t blk,
 struct ext2_xattr {
 	char *name;
 	void *value;
-	size_t value_len;
+	unsigned int value_len;
+	ext2_ino_t ea_ino;
 };
 
 struct ext2_xattr_handle {
 	errcode_t magic;
 	ext2_filsys fs;
 	struct ext2_xattr *attrs;
-	size_t length, count;
+	int capacity;
+	int count;
+	int ibody_count;
 	ext2_ino_t ino;
 	unsigned int flags;
-	int dirty;
 };
 
 static errcode_t ext2fs_xattrs_expand(struct ext2_xattr_handle *h,
@@ -226,14 +316,14 @@ static errcode_t ext2fs_xattrs_expand(struct ext2_xattr_handle *h,
 	struct ext2_xattr *new_attrs;
 	errcode_t err;
 
-	err = ext2fs_get_arrayzero(h->length + expandby,
+	err = ext2fs_get_arrayzero(h->capacity + expandby,
 				   sizeof(struct ext2_xattr), &new_attrs);
 	if (err)
 		return err;
 
-	memcpy(new_attrs, h->attrs, h->length * sizeof(struct ext2_xattr));
+	memcpy(new_attrs, h->attrs, h->capacity * sizeof(struct ext2_xattr));
 	ext2fs_free_mem(&h->attrs);
-	h->length += expandby;
+	h->capacity += expandby;
 	h->attrs = new_attrs;
 
 	return 0;
@@ -256,44 +346,6 @@ static struct ea_name_index ea_names[] = {
 	{0, NULL},
 };
 
-static int find_ea_index(char *fullname, char **name, int *index);
-
-/* Push empty attributes to the end and inlinedata to the front. */
-static int attr_compare(const void *a, const void *b)
-{
-	const struct ext2_xattr *xa = a, *xb = b;
-	char *xa_suffix, *xb_suffix;
-	int xa_idx, xb_idx;
-	int cmp;
-
-	if (xa->name == NULL)
-		return +1;
-	else if (xb->name == NULL)
-		return -1;
-	else if (!strcmp(xa->name, "system.data"))
-		return -1;
-	else if (!strcmp(xb->name, "system.data"))
-		return +1;
-
-	/*
-	 * Duplicate the kernel's sorting algorithm because xattr blocks
-	 * require sorted keys.
-	 */
-	xa_suffix = xa->name;
-	xb_suffix = xb->name;
-	xa_idx = xb_idx = 0;
-	find_ea_index(xa->name, &xa_suffix, &xa_idx);
-	find_ea_index(xb->name, &xb_suffix, &xb_idx);
-	cmp = xa_idx - xb_idx;
-	if (cmp)
-		return cmp;
-	cmp = strlen(xa_suffix) - strlen(xb_suffix);
-	if (cmp)
-		return cmp;
-	cmp = strcmp(xa_suffix, xb_suffix);
-	return cmp;
-}
-
 static const char *find_ea_prefix(int index)
 {
 	struct ea_name_index *e;
@@ -305,7 +357,7 @@ static const char *find_ea_prefix(int index)
 	return NULL;
 }
 
-static int find_ea_index(char *fullname, char **name, int *index)
+static int find_ea_index(const char *fullname, char **name, int *index)
 {
 	struct ea_name_index *e;
 
@@ -581,27 +633,21 @@ static errcode_t convert_disk_buffer_to_posix_acl(const void *value, size_t size
 	return 0;
 }
 
-
-static errcode_t write_xattrs_to_buffer(struct ext2_xattr_handle *handle,
-					struct ext2_xattr **pos,
-					void *entries_start,
-					unsigned int storage_size,
-					unsigned int value_offset_correction,
-					int write_hash)
+static errcode_t
+write_xattrs_to_buffer(ext2_filsys fs, struct ext2_xattr *attrs, int count,
+		       void *entries_start, unsigned int storage_size,
+		       unsigned int value_offset_correction, int write_hash)
 {
-	struct ext2_xattr *x = *pos;
+	struct ext2_xattr *x;
 	struct ext2_ext_attr_entry *e = entries_start;
-	char *end = (char *) entries_start + storage_size;
+	void *end = entries_start + storage_size;
 	char *shortname;
 	unsigned int entry_size, value_size;
 	int idx, ret;
+	errcode_t err;
 
 	memset(entries_start, 0, storage_size);
-	/* For all remaining x...  */
-	for (; x < handle->attrs + handle->length; x++) {
-		if (!x->name)
-			continue;
-
+	for (x = attrs; x < attrs + count; x++) {
 		/* Calculate index and shortname position */
 		shortname = x->name;
 		ret = find_ea_index(x->name, &shortname, &idx);
@@ -613,43 +659,43 @@ static errcode_t write_xattrs_to_buffer(struct ext2_xattr_handle *handle,
 		value_size = ((x->value_len + EXT2_EXT_ATTR_PAD - 1) /
 			      EXT2_EXT_ATTR_PAD) * EXT2_EXT_ATTR_PAD;
 
-		/*
-		 * Would entry collide with value?
-		 * Note that we must leave sufficient room for a (u32)0 to
-		 * mark the end of the entries.
-		 */
-		if ((char *)e + entry_size + sizeof(__u32) > end - value_size)
-			break;
-
 		/* Fill out e appropriately */
 		e->e_name_len = strlen(shortname);
 		e->e_name_index = (ret ? idx : 0);
-		e->e_value_offs = end - value_size - (char *)entries_start +
-				value_offset_correction;
-		e->e_value_block = 0;
+
 		e->e_value_size = x->value_len;
+		e->e_value_inum = x->ea_ino;
 
-		/* Store name and value */
-		end -= value_size;
+		/* Store name */
 		memcpy((char *)e + sizeof(*e), shortname, e->e_name_len);
-		memcpy(end, x->value, e->e_value_size);
+		if (x->ea_ino) {
+			e->e_value_offs = 0;
+		} else {
+			end -= value_size;
+			e->e_value_offs = end - entries_start +
+							value_offset_correction;
+			memcpy(end, x->value, e->e_value_size);
+		}
 
-		if (write_hash)
-			e->e_hash = ext2fs_ext_attr_hash_entry(e, end);
-		else
+		if (write_hash || x->ea_ino) {
+			err = ext2fs_ext_attr_hash_entry2(fs, e,
+							  x->ea_ino ? 0 : end,
+							  &e->e_hash);
+			if (err)
+				return err;
+		} else
 			e->e_hash = 0;
 
 		e = EXT2_EXT_ATTR_NEXT(e);
 		*(__u32 *)e = 0;
 	}
-	*pos = x;
-
 	return 0;
 }
 
 errcode_t ext2fs_xattrs_write(struct ext2_xattr_handle *handle)
 {
-	struct ext2_xattr *x;
+	ext2_filsys fs = handle->fs;
+	const int inode_size = EXT2_INODE_SIZE(fs->super);
 	struct ext2_inode_large *inode;
 	char *start, *block_buf = NULL;
 	struct ext2_ext_attr_header *header;
@@ -660,24 +706,23 @@ errcode_t ext2fs_xattrs_write(struct ext2_xattr_handle *handle)
 	errcode_t err;
 
 	EXT2_CHECK_MAGIC(handle, EXT2_ET_MAGIC_EA_HANDLE);
-	i = EXT2_INODE_SIZE(handle->fs->super);
+	i = inode_size;
 	if (i < sizeof(*inode))
 		i = sizeof(*inode);
 	err = ext2fs_get_memzero(i, &inode);
 	if (err)
 		return err;
 
-	err = ext2fs_read_inode_full(handle->fs, handle->ino,
-				     (struct ext2_inode *)inode,
-				     EXT2_INODE_SIZE(handle->fs->super));
+	err = ext2fs_read_inode_full(fs, handle->ino, EXT2_INODE(inode),
+				     inode_size);
 	if (err)
 		goto out;
 
 	/* If extra_isize isn't set, we need to set it now */
 	if (inode->i_extra_isize == 0 &&
-	    EXT2_INODE_SIZE(handle->fs->super) > EXT2_GOOD_OLD_INODE_SIZE) {
+	    inode_size > EXT2_GOOD_OLD_INODE_SIZE) {
 		char *p = (char *)inode;
-		size_t extra = handle->fs->super->s_want_extra_isize;
+		size_t extra = fs->super->s_want_extra_isize;
 
 		if (extra == 0)
 			extra = sizeof(__u32);
@@ -689,58 +734,45 @@ errcode_t ext2fs_xattrs_write(struct ext2_xattr_handle *handle)
 		goto out;
 	}
 
-	/*
-	 * Force the inlinedata attr to the front and the empty entries
-	 * to the end.
-	 */
-	x = handle->attrs;
-	qsort(x, handle->length, sizeof(struct ext2_xattr), attr_compare);
-
 	/* Does the inode have space for EA? */
 	if (inode->i_extra_isize < sizeof(inode->i_extra_isize) ||
-	    EXT2_INODE_SIZE(handle->fs->super) <= EXT2_GOOD_OLD_INODE_SIZE +
-						  inode->i_extra_isize +
-						  sizeof(__u32))
+	    inode_size <= EXT2_GOOD_OLD_INODE_SIZE + inode->i_extra_isize +
+								sizeof(__u32))
 		goto write_ea_block;
 
 	/* Write the inode EA */
 	ea_inode_magic = EXT2_EXT_ATTR_MAGIC;
 	memcpy(((char *) inode) + EXT2_GOOD_OLD_INODE_SIZE +
 	       inode->i_extra_isize, &ea_inode_magic, sizeof(__u32));
-	storage_size = EXT2_INODE_SIZE(handle->fs->super) -
-		EXT2_GOOD_OLD_INODE_SIZE - inode->i_extra_isize -
-		sizeof(__u32);
+	storage_size = inode_size - EXT2_GOOD_OLD_INODE_SIZE -
+				inode->i_extra_isize - sizeof(__u32);
 	start = ((char *) inode) + EXT2_GOOD_OLD_INODE_SIZE +
-		inode->i_extra_isize + sizeof(__u32);
+				inode->i_extra_isize + sizeof(__u32);
 
-	err = write_xattrs_to_buffer(handle, &x, start, storage_size, 0, 0);
+	err = write_xattrs_to_buffer(fs, handle->attrs, handle->ibody_count,
+				     start, storage_size, 0, 0);
 	if (err)
 		goto out;
-
 write_ea_block:
 	/* Are we done? */
-	if (x >= handle->attrs + handle->count)
+	if (handle->ibody_count == handle->count &&
+	    !ext2fs_file_acl_block(fs, EXT2_INODE(inode)))
 		goto skip_ea_block;
 
 	/* Write the EA block */
-	err = ext2fs_get_memzero(handle->fs->blocksize, &block_buf);
+	err = ext2fs_get_memzero(fs->blocksize, &block_buf);
 	if (err)
 		goto out;
 
-	storage_size = handle->fs->blocksize -
-		sizeof(struct ext2_ext_attr_header);
+	storage_size = fs->blocksize - sizeof(struct ext2_ext_attr_header);
 	start = block_buf + sizeof(struct ext2_ext_attr_header);
 
-	err = write_xattrs_to_buffer(handle, &x, start, storage_size,
-				     start - block_buf, 1);
+	err = write_xattrs_to_buffer(fs, handle->attrs + handle->ibody_count,
+				     handle->count - handle->ibody_count, start,
+				     storage_size, start - block_buf, 1);
 	if (err)
 		goto out2;
 
-	if (x < handle->attrs + handle->length) {
-		err = EXT2_ET_EA_NO_SPACE;
-		goto out2;
-	}
-
 	/* Write a header on the EA block */
 	header = (struct ext2_ext_attr_header *) block_buf;
 	header->h_magic = EXT2_EXT_ATTR_MAGIC;
@@ -748,31 +780,28 @@ write_ea_block:
 	header->h_blocks = 1;
 
 	/* Get a new block for writing */
-	err = prep_ea_block_for_write(handle->fs, handle->ino, inode);
+	err = prep_ea_block_for_write(fs, handle->ino, inode);
 	if (err)
 		goto out2;
 
 	/* Finally, write the new EA block */
-	blk = ext2fs_file_acl_block(handle->fs,
-				    (struct ext2_inode *)inode);
-	err = ext2fs_write_ext_attr3(handle->fs, blk, block_buf,
-				     handle->ino);
+	blk = ext2fs_file_acl_block(fs, EXT2_INODE(inode));
+	err = ext2fs_write_ext_attr3(fs, blk, block_buf, handle->ino);
 	if (err)
 		goto out2;
 
 skip_ea_block:
-	blk = ext2fs_file_acl_block(handle->fs, (struct ext2_inode *)inode);
+	blk = ext2fs_file_acl_block(fs, (struct ext2_inode *)inode);
 	if (!block_buf && blk) {
 		/* xattrs shrunk, free the block */
-		err = ext2fs_free_ext_attr(handle->fs, handle->ino, inode);
+		err = ext2fs_free_ext_attr(fs, handle->ino, inode);
 		if (err)
 			goto out;
 	}
 
 	/* Write the inode */
-	err = ext2fs_write_inode_full(handle->fs, handle->ino,
-				      (struct ext2_inode *)inode,
-				      EXT2_INODE_SIZE(handle->fs->super));
+	err = ext2fs_write_inode_full(fs, handle->ino, EXT2_INODE(inode),
+				      inode_size);
 	if (err)
 		goto out2;
 
@@ -780,15 +809,14 @@ out2:
 	ext2fs_free_mem(&block_buf);
 out:
 	ext2fs_free_mem(&inode);
-	handle->dirty = 0;
 	return err;
 }
 
 static errcode_t read_xattrs_from_buffer(struct ext2_xattr_handle *handle,
+					 struct ext2_inode_large *inode,
 					 struct ext2_ext_attr_entry *entries,
 					 unsigned int storage_size,
-					 char *value_start,
-					 size_t *nr_read)
+					 char *value_start)
 {
 	struct ext2_xattr *x;
 	struct ext2_ext_attr_entry *entry, *end;
@@ -798,10 +826,6 @@ static errcode_t read_xattrs_from_buffer(struct ext2_xattr_handle *handle,
 	unsigned int values_size = storage_size +
 			((char *)entries - value_start);
 
-	x = handle->attrs;
-	while (x->name)
-		x++;
-
 	/* find the end */
 	end = entries;
 	remain = storage_size;
@@ -824,7 +848,15 @@ static errcode_t read_xattrs_from_buffer(struct ext2_xattr_handle *handle,
 	remain = storage_size;
 	while (remain >= sizeof(struct ext2_ext_attr_entry) &&
 	       !EXT2_EXT_IS_LAST_ENTRY(entry)) {
-		__u32 hash;
+
+		/* Allocate space for more attrs? */
+		if (handle->count == handle->capacity) {
+			err = ext2fs_xattrs_expand(handle, 4);
+			if (err)
+				return err;
+		}
+
+		x = handle->attrs + handle->count;
 
 		/* header eats this space */
 		remain -= sizeof(struct ext2_ext_attr_entry);
@@ -832,40 +864,7 @@ static errcode_t read_xattrs_from_buffer(struct ext2_xattr_handle *handle,
 		/* attribute len eats this space */
 		remain -= EXT2_EXT_ATTR_SIZE(entry->e_name_len);
 
-		/* check value size */
-		if (entry->e_value_size > remain)
-			return EXT2_ET_EA_BAD_VALUE_SIZE;
-
-		if (entry->e_value_offs + entry->e_value_size > values_size)
-			return EXT2_ET_EA_BAD_VALUE_OFFSET;
-
-		if (entry->e_value_size > 0 &&
-		    value_start + entry->e_value_offs <
-		    (char *)end + sizeof(__u32))
-			return EXT2_ET_EA_BAD_VALUE_OFFSET;
-
-		/* e_value_block must be 0 in inode's ea */
-		if (entry->e_value_block != 0)
-			return EXT2_ET_BAD_EA_BLOCK_NUM;
-
-		hash = ext2fs_ext_attr_hash_entry(entry, value_start +
-							 entry->e_value_offs);
-
-		/* e_hash may be 0 in older inode's ea */
-		if (entry->e_hash != 0 && entry->e_hash != hash)
-			return EXT2_ET_BAD_EA_HASH;
-
-		remain -= entry->e_value_size;
-
-		/* Allocate space for more attrs? */
-		if (x == handle->attrs + handle->length) {
-			err = ext2fs_xattrs_expand(handle, 4);
-			if (err)
-				return err;
-			x = handle->attrs + handle->length - 4;
-		}
-
-		/* Extract name/value */
+		/* Extract name */
 		prefix = find_ea_prefix(entry->e_name_index);
 		prefix_len = (prefix ? strlen(prefix) : 0);
 		err = ext2fs_get_memzero(entry->e_name_len + prefix_len + 1,
@@ -879,14 +878,96 @@ static errcode_t read_xattrs_from_buffer(struct ext2_xattr_handle *handle,
 			       (char *)entry + sizeof(*entry),
 			       entry->e_name_len);
 
-		err = ext2fs_get_mem(entry->e_value_size, &x->value);
-		if (err)
-			return err;
+		/* Check & copy value */
+		if (!ext2fs_has_feature_ea_inode(handle->fs->super) &&
+		    entry->e_value_inum != 0)
+			return EXT2_ET_BAD_EA_BLOCK_NUM;
+
+		if (entry->e_value_inum == 0) {
+			if (entry->e_value_size > remain)
+				return EXT2_ET_EA_BAD_VALUE_SIZE;
+
+			if (entry->e_value_offs + entry->e_value_size > values_size)
+				return EXT2_ET_EA_BAD_VALUE_OFFSET;
+
+			if (entry->e_value_size > 0 &&
+			    value_start + entry->e_value_offs <
+			    (char *)end + sizeof(__u32))
+				return EXT2_ET_EA_BAD_VALUE_OFFSET;
+
+			remain -= entry->e_value_size;
+
+			err = ext2fs_get_mem(entry->e_value_size, &x->value);
+			if (err)
+				return err;
+			memcpy(x->value, value_start + entry->e_value_offs,
+			       entry->e_value_size);
+		} else {
+			struct ext2_inode *ea_inode;
+			ext2_file_t ea_file;
+
+			if (entry->e_value_offs != 0)
+				return EXT2_ET_EA_BAD_VALUE_OFFSET;
+
+			if (entry->e_value_size > (64 * 1024))
+				return EXT2_ET_EA_BAD_VALUE_SIZE;
+
+			err = ext2fs_get_mem(entry->e_value_size, &x->value);
+			if (err)
+				return err;
+
+			err = ext2fs_file_open(handle->fs, entry->e_value_inum,
+					       0, &ea_file);
+			if (err)
+				return err;
+
+			ea_inode = ext2fs_file_get_inode(ea_file);
+			if ((ea_inode->i_flags & EXT4_INLINE_DATA_FL) ||
+			    !(ea_inode->i_flags & EXT4_EA_INODE_FL) ||
+			    ea_inode->i_links_count == 0)
+				err = EXT2_ET_EA_INODE_CORRUPTED;
+			else if (ext2fs_file_get_size(ea_file) !=
+			    entry->e_value_size)
+				err = EXT2_ET_EA_BAD_VALUE_SIZE;
+			else
+				err = ext2fs_file_read(ea_file, x->value,
+						       entry->e_value_size, 0);
+			ext2fs_file_close(ea_file);
+			if (err)
+				return err;
+		}
+
+		x->ea_ino = entry->e_value_inum;
 		x->value_len = entry->e_value_size;
-		memcpy(x->value, value_start + entry->e_value_offs,
-		       entry->e_value_size);
-		x++;
-		(*nr_read)++;
+
+		/* e_hash may be 0 in older inode's ea */
+		if (entry->e_hash != 0) {
+			__u32 hash;
+			void *data = (entry->e_value_inum != 0) ?
+					0 : value_start + entry->e_value_offs;
+
+			err = ext2fs_ext_attr_hash_entry2(handle->fs, entry,
+							  data, &hash);
+			if (err)
+				return err;
+			if (entry->e_hash != hash) {
+				struct ext2_inode child;
+
+				/* Check whether this is an old Lustre-style
+				 * ea_inode reference.
+				 */
+				err = ext2fs_read_inode(handle->fs,
+							entry->e_value_inum,
+							&child);
+				if (err)
+					return err;
+				if (child.i_mtime != handle->ino ||
+				    child.i_generation != inode->i_generation)
+					return EXT2_ET_BAD_EA_HASH;
+			}
+		}
+
+		handle->count++;
 		entry = EXT2_EXT_ATTR_NEXT(entry);
 	}
 
@@ -898,13 +979,14 @@ static void xattrs_free_keys(struct ext2_xattr_handle *h)
 	struct ext2_xattr *a = h->attrs;
 	size_t i;
 
-	for (i = 0; i < h->length; i++) {
+	for (i = 0; i < h->capacity; i++) {
 		if (a[i].name)
 			ext2fs_free_mem(&a[i].name);
 		if (a[i].value)
 			ext2fs_free_mem(&a[i].value);
 	}
 	h->count = 0;
+	h->ibody_count = 0;
 }
 
 errcode_t ext2fs_xattrs_read(struct ext2_xattr_handle *handle)
@@ -955,11 +1037,13 @@ errcode_t ext2fs_xattrs_read(struct ext2_xattr_handle *handle)
 		start = ((char *) inode) + EXT2_GOOD_OLD_INODE_SIZE +
 			inode->i_extra_isize + sizeof(__u32);
 
-		err = read_xattrs_from_buffer(handle,
-			(struct ext2_ext_attr_entry *) start, storage_size,
-					      start, &handle->count);
+		err = read_xattrs_from_buffer(handle, inode,
+					(struct ext2_ext_attr_entry *) start,
+					storage_size, start);
 		if (err)
 			goto out;
+
+		handle->ibody_count = handle->count;
 	}
 
 read_ea_block:
@@ -992,9 +1076,9 @@ read_ea_block:
 		storage_size = handle->fs->blocksize -
 			sizeof(struct ext2_ext_attr_header);
 		start = block_buf + sizeof(struct ext2_ext_attr_header);
-		err = read_xattrs_from_buffer(handle,
-			(struct ext2_ext_attr_entry *) start, storage_size,
-					      block_buf, &handle->count);
+		err = read_xattrs_from_buffer(handle, inode,
+					(struct ext2_ext_attr_entry *) start,
+					storage_size, block_buf);
 		if (err)
 			goto out3;
 
@@ -1018,20 +1102,20 @@ errcode_t ext2fs_xattrs_iterate(struct ext2_xattr_handle *h,
 				void *data)
 {
 	struct ext2_xattr *x;
+	int dirty = 0;
 	int ret;
 
 	EXT2_CHECK_MAGIC(h, EXT2_ET_MAGIC_EA_HANDLE);
-	for (x = h->attrs; x < h->attrs + h->length; x++) {
-		if (!x->name)
-			continue;
-
+	for (x = h->attrs; x < h->attrs + h->count; x++) {
 		ret = func(x->name, x->value, x->value_len, data);
 		if (ret & XATTR_CHANGED)
-			h->dirty = 1;
+			dirty = 1;
 		if (ret & XATTR_ABORT)
-			return 0;
+			break;
 	}
 
+	if (dirty)
+		return ext2fs_xattrs_write(h);
 	return 0;
 }
 
@@ -1043,8 +1127,8 @@ errcode_t ext2fs_xattr_get(struct ext2_xattr_handle *h, const char *key,
 	errcode_t err;
 
 	EXT2_CHECK_MAGIC(h, EXT2_ET_MAGIC_EA_HANDLE);
-	for (x = h->attrs; x < h->attrs + h->length; x++) {
-		if (!x->name || strcmp(x->name, key))
+	for (x = h->attrs; x < h->attrs + h->count; x++) {
+		if (strcmp(x->name, key))
 			continue;
 
 		if (!(h->flags & XATTR_HANDLE_FLAG_RAW) &&
@@ -1107,7 +1191,7 @@ errcode_t ext2fs_xattr_inode_max_size(ext2_filsys fs, ext2_ino_t ino,
 			inode->i_extra_isize + sizeof(__u32);
 		entry = (struct ext2_ext_attr_entry *) start;
 		while (!EXT2_EXT_IS_LAST_ENTRY(entry)) {
-			if (!entry->e_value_block && entry->e_value_size) {
+			if (!entry->e_value_inum && entry->e_value_size) {
 				unsigned int offs = entry->e_value_offs;
 				if (offs < minoff)
 					minoff = offs;
@@ -1127,101 +1211,448 @@ out:
 	return err;
 }
 
-errcode_t ext2fs_xattr_set(struct ext2_xattr_handle *handle,
-			   const char *key,
-			   const void *value,
-			   size_t value_len)
+static errcode_t xattr_create_ea_inode(ext2_filsys fs, const void *value,
+				       size_t value_len, ext2_ino_t *ea_ino)
 {
-	struct ext2_xattr *x, *last_empty;
-	char *new_value;
-	errcode_t err;
+	struct ext2_inode inode;
+	ext2_ino_t ino;
+	ext2_file_t file;
+	__u32 hash;
+	errcode_t ret;
 
-	EXT2_CHECK_MAGIC(handle, EXT2_ET_MAGIC_EA_HANDLE);
-	last_empty = NULL;
+	ret = ext2fs_new_inode(fs, 0, 0, 0, &ino);
+	if (ret)
+		return ret;
 
-	err = ext2fs_get_mem(value_len, &new_value);
-	if (err)
-		return err;
-	if (!(handle->flags & XATTR_HANDLE_FLAG_RAW) &&
-	    ((strcmp(key, "system.posix_acl_default") == 0) ||
-	     (strcmp(key, "system.posix_acl_access") == 0))) {
-		err = convert_posix_acl_to_disk_buffer(value, value_len,
-						       new_value, &value_len);
-		if (err)
-			goto errout;
-	} else
-		memcpy(new_value, value, value_len);
+	memset(&inode, 0, sizeof(inode));
+	inode.i_flags |= EXT4_EA_INODE_FL;
+	if (ext2fs_has_feature_extents(fs->super))
+		inode.i_flags |= EXT4_EXTENTS_FL;
+	inode.i_size = 0;
+	inode.i_mode = LINUX_S_IFREG | 0600;
+	inode.i_links_count = 1;
+	ret = ext2fs_write_new_inode(fs, ino, &inode);
+	if (ret)
+		return ret;
+	/*
+	 * ref_count and hash utilize inode's i_*time fields.
+	 * ext2fs_write_new_inode() call above initializes these fields with
+	 * current time. That's why ref count and hash updates are done
+	 * separately below.
+	 */
+	ext2fs_set_ea_inode_ref(&inode, 1);
+	hash = ext2fs_crc32c_le(fs->csum_seed, value, value_len);
+	ext2fs_set_ea_inode_hash(&inode, hash);
 
-	for (x = handle->attrs; x < handle->attrs + handle->length; x++) {
-		if (!x->name) {
-			last_empty = x;
-			continue;
-		}
+	ret = ext2fs_write_inode(fs, ino, &inode);
+	if (ret)
+		return ret;
 
-		/* Replace xattr */
-		if (strcmp(x->name, key) == 0) {
-			ext2fs_free_mem(&x->value);
-			x->value = new_value;
-			x->value_len = value_len;
-			handle->dirty = 1;
-			return 0;
-		}
+	ret = ext2fs_file_open(fs, ino, EXT2_FILE_WRITE, &file);
+	if (ret)
+		return ret;
+	ret = ext2fs_file_write(file, value, value_len, NULL);
+	ext2fs_file_close(file);
+	if (ret)
+		return ret;
+
+	ext2fs_inode_alloc_stats2(fs, ino, 1 /* inuse */, 0 /* isdir */);
+
+	*ea_ino = ino;
+	return 0;
+}
+
+static errcode_t xattr_inode_dec_ref(ext2_filsys fs, ext2_ino_t ino)
+{
+	struct ext2_inode_large inode;
+	__u64 ref_count;
+	errcode_t ret;
+
+	ret = ext2fs_read_inode_full(fs, ino, (struct ext2_inode *)&inode,
+				     sizeof(inode));
+	if (ret)
+		goto out;
+
+	ref_count = ext2fs_get_ea_inode_ref(EXT2_INODE(&inode));
+	ref_count--;
+	ext2fs_set_ea_inode_ref(EXT2_INODE(&inode), ref_count);
+
+	if (ref_count)
+		goto write_out;
+
+	inode.i_links_count = 0;
+	inode.i_dtime = fs->now ? fs->now : time(0);
+
+	ret = ext2fs_free_ext_attr(fs, ino, &inode);
+	if (ret)
+		goto write_out;
+
+	if (ext2fs_inode_has_valid_blocks2(fs, (struct ext2_inode *)&inode)) {
+		ret = ext2fs_punch(fs, ino, (struct ext2_inode *)&inode, NULL,
+				   0, ~0ULL);
+		if (ret)
+			goto out;
 	}
 
-	/* Add attr to empty slot */
-	if (last_empty) {
-		err = ext2fs_get_mem(strlen(key) + 1, &last_empty->name);
-		if (err)
-			goto errout;
-		strcpy(last_empty->name, key);
-		last_empty->value = new_value;
-		last_empty->value_len = value_len;
-		handle->dirty = 1;
-		handle->count++;
+	ext2fs_inode_alloc_stats2(fs, ino, -1 /* inuse */, 0 /* is_dir */);
+
+write_out:
+	ret = ext2fs_write_inode_full(fs, ino, (struct ext2_inode *)&inode,
+				      sizeof(inode));
+out:
+	return ret;
+}
+
+static errcode_t xattr_update_entry(ext2_filsys fs, struct ext2_xattr *x,
+				    const char *name, const void *value,
+				    size_t value_len, int in_inode)
+{
+	ext2_ino_t ea_ino = 0;
+	void *new_value = NULL;
+	char *new_name = NULL;
+	int name_len;
+	errcode_t ret;
+
+	if (!x->name) {
+		name_len = strlen(name);
+		ret = ext2fs_get_mem(name_len + 1, &new_name);
+		if (ret)
+			goto fail;
+		memcpy(new_name, name, name_len + 1);
+	}
+
+	ret = ext2fs_get_mem(value_len, &new_value);
+	if (ret)
+		goto fail;
+	memcpy(new_value, value, value_len);
+
+	if (in_inode) {
+		ret = xattr_create_ea_inode(fs, value, value_len, &ea_ino);
+		if (ret)
+			goto fail;
+	}
+
+	if (x->ea_ino) {
+		ret = xattr_inode_dec_ref(fs, x->ea_ino);
+		if (ret)
+			goto fail;
+	}
+
+	if (!x->name)
+		x->name = new_name;
+
+	if (x->value)
+		ext2fs_free_mem(&x->value);
+	x->value = new_value;
+	x->value_len = value_len;
+	x->ea_ino = ea_ino;
+	return 0;
+fail:
+	if (new_name)
+		ext2fs_free_mem(&new_name);
+	if (new_value)
+		ext2fs_free_mem(&new_value);
+	if (ea_ino)
+		xattr_inode_dec_ref(fs, ea_ino);
+	return ret;
+}
+
+static int xattr_find_position(struct ext2_xattr *attrs, int count,
+			       const char *name)
+{
+	struct ext2_xattr *x;
+	int i;
+	char *shortname, *x_shortname;
+	int name_idx, x_name_idx;
+	int shortname_len, x_shortname_len;
+
+	find_ea_index(name, &shortname, &name_idx);
+	shortname_len = strlen(shortname);
+
+	for (i = 0, x = attrs; i < count; i++, x++) {
+		find_ea_index(x->name, &x_shortname, &x_name_idx);
+		if (name_idx < x_name_idx)
+			break;
+		if (name_idx > x_name_idx)
+			continue;
+
+		x_shortname_len = strlen(x_shortname);
+		if (shortname_len < x_shortname_len)
+			break;
+		if (shortname_len > x_shortname_len)
+			continue;
+
+		if (memcmp(shortname, x_shortname, shortname_len) <= 0)
+			break;
+	}
+	return i;
+}
+
+static errcode_t xattr_array_update(struct ext2_xattr_handle *h,
+				    const char *name,
+				    const void *value, size_t value_len,
+				    int ibody_free, int block_free,
+				    int old_idx, int in_inode)
+{
+	struct ext2_xattr tmp;
+	int add_to_ibody;
+	int needed;
+	int name_len, name_idx;
+	char *shortname;
+	int new_idx;
+	int ret;
+
+	find_ea_index(name, &shortname, &name_idx);
+	name_len = strlen(shortname);
+
+	needed = EXT2_EXT_ATTR_LEN(name_len);
+	if (!in_inode)
+		needed += EXT2_EXT_ATTR_SIZE(value_len);
+
+	if (old_idx >= 0 && old_idx < h->ibody_count) {
+		ibody_free += EXT2_EXT_ATTR_LEN(name_len);
+		if (!h->attrs[old_idx].ea_ino)
+			ibody_free += EXT2_EXT_ATTR_SIZE(
+						h->attrs[old_idx].value_len);
+	}
+
+	if (needed <= ibody_free) {
+		if (old_idx < 0) {
+			new_idx = h->ibody_count;
+			add_to_ibody = 1;
+			goto add_new;
+		}
+
+		/* Update the existing entry. */
+		ret = xattr_update_entry(h->fs, &h->attrs[old_idx], name,
+					 value, value_len, in_inode);
+		if (ret)
+			return ret;
+		if (h->ibody_count <= old_idx) {
+			/* Move entry from block to the end of ibody. */
+			tmp = h->attrs[old_idx];
+			memmove(h->attrs + h->ibody_count + 1,
+				h->attrs + h->ibody_count,
+				(old_idx - h->ibody_count) * sizeof(*h->attrs));
+			h->attrs[h->ibody_count] = tmp;
+			h->ibody_count++;
+		}
 		return 0;
 	}
 
-	/* Expand array, append slot */
-	err = ext2fs_xattrs_expand(handle, 4);
-	if (err)
-		goto errout;
+	if (h->ibody_count <= old_idx) {
+		block_free += EXT2_EXT_ATTR_LEN(name_len);
+		if (!h->attrs[old_idx].ea_ino)
+			block_free +=
+				EXT2_EXT_ATTR_SIZE(h->attrs[old_idx].value_len);
+	}
 
-	x = handle->attrs + handle->length - 4;
-	err = ext2fs_get_mem(strlen(key) + 1, &x->name);
-	if (err)
-		goto errout;
-	strcpy(x->name, key);
+	if (needed > block_free)
+		return EXT2_ET_EA_NO_SPACE;
 
-	err = ext2fs_get_mem(value_len, &x->value);
-	if (err)
-		goto errout;
-	memcpy(x->value, value, value_len);
-	x->value_len = value_len;
-	handle->dirty = 1;
-	handle->count++;
+	if (old_idx >= 0) {
+		/* Update the existing entry. */
+		ret = xattr_update_entry(h->fs, &h->attrs[old_idx], name,
+					 value, value_len, in_inode);
+		if (ret)
+			return ret;
+		if (old_idx < h->ibody_count) {
+			/*
+			 * Move entry from ibody to the block. Note that
+			 * entries in the block are sorted.
+			 */
+			new_idx = xattr_find_position(h->attrs + h->ibody_count,
+				h->count - h->ibody_count, name);
+			new_idx += h->ibody_count - 1;
+			tmp = h->attrs[old_idx];
+			memmove(h->attrs + old_idx, h->attrs + old_idx + 1,
+				(new_idx - old_idx) * sizeof(*h->attrs));
+			h->attrs[new_idx] = tmp;
+			h->ibody_count--;
+		}
+		return 0;
+	}
+
+	new_idx = xattr_find_position(h->attrs + h->ibody_count,
+				      h->count - h->ibody_count, name);
+	new_idx += h->ibody_count;
+	add_to_ibody = 0;
+
+add_new:
+	if (h->count == h->capacity) {
+		ret = ext2fs_xattrs_expand(h, 4);
+		if (ret)
+			return ret;
+	}
+
+	ret = xattr_update_entry(h->fs, &h->attrs[h->count], name, value,
+				 value_len, in_inode);
+	if (ret)
+		return ret;
+
+	tmp = h->attrs[h->count];
+	memmove(h->attrs + new_idx + 1, h->attrs + new_idx,
+		(h->count - new_idx)*sizeof(*h->attrs));
+	h->attrs[new_idx] = tmp;
+	if (add_to_ibody)
+		h->ibody_count++;
+	h->count++;
 	return 0;
-errout:
+}
+
+static int space_used(struct ext2_xattr *attrs, int count)
+{
+	int total = 0;
+	struct ext2_xattr *x;
+	char *shortname;
+	int i, len, name_idx;
+
+	for (i = 0, x = attrs; i < count; i++, x++) {
+		find_ea_index(x->name, &shortname, &name_idx);
+		len = strlen(shortname);
+		total += EXT2_EXT_ATTR_LEN(len);
+		if (!x->ea_ino)
+			total += EXT2_EXT_ATTR_SIZE(x->value_len);
+	}
+	return total;
+}
+
+/*
+ * The minimum size of EA value when you start storing it in an external inode
+ * size of block - size of header - size of 1 entry - 4 null bytes
+ */
+#define EXT4_XATTR_MIN_LARGE_EA_SIZE(b)	\
+	((b) - EXT2_EXT_ATTR_LEN(3) - sizeof(struct ext2_ext_attr_header) - 4)
+
+errcode_t ext2fs_xattr_set(struct ext2_xattr_handle *h,
+			   const char *name,
+			   const void *value,
+			   size_t value_len)
+{
+	ext2_filsys fs = h->fs;
+	const int inode_size = EXT2_INODE_SIZE(fs->super);
+	struct ext2_inode_large *inode = NULL;
+	struct ext2_xattr *x;
+	char *new_value;
+	int ibody_free, block_free;
+	int in_inode = 0;
+	int old_idx = -1;
+	int extra_isize;
+	errcode_t ret;
+
+	EXT2_CHECK_MAGIC(h, EXT2_ET_MAGIC_EA_HANDLE);
+
+	ret = ext2fs_get_mem(value_len, &new_value);
+	if (ret)
+		return ret;
+	if (!(h->flags & XATTR_HANDLE_FLAG_RAW) &&
+	    ((strcmp(name, "system.posix_acl_default") == 0) ||
+	     (strcmp(name, "system.posix_acl_access") == 0))) {
+		ret = convert_posix_acl_to_disk_buffer(value, value_len,
+						       new_value, &value_len);
+		if (ret)
+			goto out;
+	} else
+		memcpy(new_value, value, value_len);
+
+	/* Imitate kernel behavior by skipping update if value is the same. */
+	for (x = h->attrs; x < h->attrs + h->count; x++) {
+		if (!strcmp(x->name, name)) {
+			if (!x->ea_ino && x->value_len == value_len &&
+			    !memcmp(x->value, new_value, value_len)) {
+				ret = 0;
+				goto out;
+			}
+			old_idx = x - h->attrs;
+			break;
+		}
+	}
+
+	ret = ext2fs_get_memzero(inode_size, &inode);
+	if (ret)
+		goto out;
+	ret = ext2fs_read_inode_full(fs, h->ino,
+				     (struct ext2_inode *)inode,
+				     inode_size);
+	if (ret)
+		goto out;
+	if (inode_size > EXT2_GOOD_OLD_INODE_SIZE) {
+		extra_isize = inode->i_extra_isize;
+		if (extra_isize == 0) {
+			extra_isize = fs->super->s_want_extra_isize;
+			if (extra_isize == 0)
+				extra_isize = sizeof(__u32);
+		}
+		ibody_free = inode_size - EXT2_GOOD_OLD_INODE_SIZE;
+		ibody_free -= extra_isize;
+		/* Extended attribute magic and final null entry. */
+		ibody_free -= sizeof(__u32) * 2;
+		ibody_free -= space_used(h->attrs, h->ibody_count);
+	} else
+		ibody_free = 0;
+
+	/* Inline data can only go to ibody. */
+	if (strcmp(name, "system.data") == 0) {
+		if (h->ibody_count <= old_idx) {
+			ret = EXT2_ET_FILESYSTEM_CORRUPTED;
+			goto out;
+		}
+		ret = xattr_array_update(h, name, value, value_len, ibody_free,
+					 0 /* block_free */, old_idx,
+					 0 /* in_inode */);
+		if (ret)
+			goto out;
+		goto write_out;
+	}
+
+	block_free = fs->blocksize;
+	block_free -= sizeof(struct ext2_ext_attr_header);
+	/* Final null entry. */
+	block_free -= sizeof(__u32);
+	block_free -= space_used(h->attrs + h->ibody_count,
+				 h->count - h->ibody_count);
+
+	if (ext2fs_has_feature_ea_inode(fs->super) &&
+	    value_len > EXT4_XATTR_MIN_LARGE_EA_SIZE(fs->blocksize))
+		in_inode = 1;
+
+	ret = xattr_array_update(h, name, value, value_len, ibody_free,
+				 block_free, old_idx, in_inode);
+	if (ret == EXT2_ET_EA_NO_SPACE && !in_inode &&
+	    ext2fs_has_feature_ea_inode(fs->super))
+		ret = xattr_array_update(h, name, value, value_len, ibody_free,
+				 block_free, old_idx, 1 /* in_inode */);
+	if (ret)
+		goto out;
+
+write_out:
+	ret = ext2fs_xattrs_write(h);
+out:
+	if (inode)
+		ext2fs_free_mem(&inode);
 	ext2fs_free_mem(&new_value);
-	return err;
+	return ret;
 }
 
 errcode_t ext2fs_xattr_remove(struct ext2_xattr_handle *handle,
 			      const char *key)
 {
 	struct ext2_xattr *x;
+	struct ext2_xattr *end = handle->attrs + handle->count;
 
 	EXT2_CHECK_MAGIC(handle, EXT2_ET_MAGIC_EA_HANDLE);
-	for (x = handle->attrs; x < handle->attrs + handle->length; x++) {
-		if (!x->name)
-			continue;
-
+	for (x = handle->attrs; x < end; x++) {
 		if (strcmp(x->name, key) == 0) {
 			ext2fs_free_mem(&x->name);
 			ext2fs_free_mem(&x->value);
-			x->value_len = 0;
-			handle->dirty = 1;
+			if (x->ea_ino)
+				xattr_inode_dec_ref(handle->fs, x->ea_ino);
+			memmove(x, x + 1, (end - x - 1)*sizeof(*x));
+			memset(end - 1, 0, sizeof(*end));
+			if (x < handle->attrs + handle->ibody_count)
+				handle->ibody_count--;
 			handle->count--;
-			return 0;
+			return ext2fs_xattrs_write(handle);
 		}
 	}
 
@@ -1244,8 +1675,8 @@ errcode_t ext2fs_xattrs_open(ext2_filsys fs, ext2_ino_t ino,
 		return err;
 
 	h->magic = EXT2_ET_MAGIC_EA_HANDLE;
-	h->length = 4;
-	err = ext2fs_get_arrayzero(h->length, sizeof(struct ext2_xattr),
+	h->capacity = 4;
+	err = ext2fs_get_arrayzero(h->capacity, sizeof(struct ext2_xattr),
 				   &h->attrs);
 	if (err) {
 		ext2fs_free_mem(&h);
@@ -1261,15 +1692,8 @@ errcode_t ext2fs_xattrs_open(ext2_filsys fs, ext2_ino_t ino,
 errcode_t ext2fs_xattrs_close(struct ext2_xattr_handle **handle)
 {
 	struct ext2_xattr_handle *h = *handle;
-	errcode_t err;
 
 	EXT2_CHECK_MAGIC(h, EXT2_ET_MAGIC_EA_HANDLE);
-	if (h->dirty) {
-		err = ext2fs_xattrs_write(h);
-		if (err)
-			return err;
-	}
-
 	xattrs_free_keys(h);
 	ext2fs_free_mem(&h->attrs);
 	ext2fs_free_mem(handle);
